# safeinit.tcl --
#
# This code runs in a master to manage a safe slave with Safe Tcl.
# See the safe.n man page for details.
#
# Copyright (c) 1996 Sun Microsystems, Inc.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# SCCS: @(#) safeinit.tcl 1.19 96/12/11 16:42:41

# The following code ensures that tcl_PolicyPath has a reasonable initial
# value:

if {![info exists tcl_policyPath]} {
    set tcl_policyPath {}		;# NOTE: Must be a list of paths!
    if {"$tcl_platform(platform)" == "unix"} {
	catch {set tcl_policyPath $env(TCL_POLICY_PATH)}
    }
}

# This procedure creates a safe slave, initializes it with the
# safe base and installs the aliases for the security policy mechanism.

proc tcl_safeCreateInterp {slave} {
    upvar #0 tclSafe$slave state
    global tcl_library tk_library

    # Create the slave

    interp create -safe $slave

    # These aliases let the slave load files to define new commands

    interp alias $slave source {} tclSafeAliasSource $slave
    interp alias $slave load {} tclSafeAliasLoad $slave

    # This alias lets the slave have access to a subset of the 'file'
    # command functionality.
    tclAliasSubset $slave file file dir.* join root.* ext.* tail \
	path.* split

    # This alias interposes on the 'exit' command and cleanly terminates
    # the slave.
    interp alias $slave exit {} tcl_safeDeleteInterp $slave

    # Install a few library routines
    foreach proc {tclPkgSetup unknown auto_load} {
	catch {$proc}	;# force auto load
	interp eval $slave \
	    [list proc $proc [info args $proc] [info body $proc]]
    }
    # Loading packages into slaves is handled by their master.
    # This is overloaded to deal with regular packages and security policies

    interp alias $slave tclPkgUnknown {} tclSafeAliasPkgUnknown $slave
    interp eval $slave {package unknown tclPkgUnknown}

    # We need a helper procedure to define a $dir variable and then
    # do a source of the pkgIndex.tcl file
    interp eval $slave [list proc tclPkgSource {dir file} {source $file }]

    # Let the slave inherit a few variables
    foreach varName \
	{tcl_library tcl_version tcl_patchLevel tcl_platform(platform)} {
	upvar #0 $varName var
	interp eval $slave [list set $varName $var]
    }

    # Other variables are predefined with set values
    foreach {varName value} {
	    auto_noexec 1
	    auto_path {}
	    errorCode {}
	    errorInfo {}
	    env() {}
	    argv0 {}
	    argv {}
	    argc 0
	    tcl_interactive 0
	    } {
	interp eval $slave [list set $varName $value]
    }

    # If we have Tk, make the slave have the same library as us:

    if {[info exists tk_library]} {
        interp eval $slave [list set tk_library $tk_library]
    }

    # Create the default auto-path for the slave, if there is not one yet:
    if {![info exists state(auto_path)]} {
	set state(auto_path) [list $tcl_library]
	if {[info exists tk_library]} {
	    lappend state(auto_path) $tk_library
	}
    }

    # Stub out auto-exec mechanism in slave
    interp eval $slave [list proc auto_execok {name} {return {}}]

    return $slave
}

# This procedure deletes a safe slave managed by Safe Tcl and
# cleans up associated state:

proc tcl_safeDeleteInterp {slave args} {
    upvar #0 tclSafe$slave state

    # If the slave has a policy loaded, clean it up now.
    if {[info exists state(policyLoaded)]} {
	set policy $state(policyLoaded)
	set proc ${policy}_PolicyCleanup
	if {[string compare [info proc $proc] $proc] == 0} {
	    $proc $slave
	}
    }

    # Discard the global array of state associated with the slave, and
    # delete the interpreter.
    catch {unset state}
    catch {interp delete $slave}

    return
}

# This procedure manages the per-slave auto-path for finding packages.

proc tcl_safeAutoPath {slave args} {
    upvar #0 tclSafe$slave state
    if {[llength $args] == 0} {
	return $state(auto_path)
    } else {
	set state(auto_path) [lindex $args 0]
    }
}

# This procedure manages the global security policy search path

proc tcl_safePolicyPath args {
    global tcl_policyPath

    if {[llength $args] == 0} {
	return $tcl_policyPath
    } else {
	set tcl_policyPath [lindex $args 0]
    }
}

# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------

# tclSafeAliasSource is the target of the "source" alias in safe interpreters.

proc tclSafeAliasSource {slave args} {
    upvar #0 tclSafe$slave state
    global auto_path

    if {[llength $args] == 2} {
	if {[string compare "-rsrc" [lindex $args 0]] != 0} {
	    return -code error "incorrect arguments to source"
	}
	if {[catch {interp invokehidden $slave source -rsrc [lindex $args 1]} msg]} {
	    return -code error $msg
	}
    } else {
        # Use the main auto_path, or the per-slave auto_path
        set path $auto_path
        catch {set path $state(auto_path)}

	set file [lindex $args 0]
	if {[catch {tclFileInPath $file $path $slave} msg]} {
	    return -code error "permission denied"
	}
	if {[catch {interp invokehidden $slave source $file} msg]} {
	    return -code error $msg
	}
    }
    return $msg
}

# tclSafeAliasLoad is the target of the "load" alias in safe interpreters.

proc tclSafeAliasLoad {slave file args} {
    upvar #0 tclSafe$slave state
    global auto_path

    if {[llength $args] == 2} {
	# Trying to load into another interpreter
	# Allow this for a child of the slave, or itself
	set other [lindex $args 1]
	foreach x $slave y $other {
	    if {[string length $x] == 0} {
		break
	    } elseif {[string compare $x $y] != 0} {
		return -code error "permission denied"
	    }
	}
	set slave $other
    }

    # Use the main auto_path, or the per-slave auto_path
    set path $auto_path
    catch {set path $state(auto_path)}

    if {[string length $file] && \
		[catch {tclFileInPath $file $path $slave} msg]} {
	return -code error "permission denied"
    }
    if {[catch {
	switch [llength $args] {
	    0 {
		interp invokehidden $slave load $file
	    }
	    1 -
	    2 {
		interp invokehidden $slave load $file [lindex $args 0]
	    }
	    default {
		error "too many arguments to load"
	    }
	}
    } msg]} {
	return -code error $msg
    }
    return $msg
}

# tclFileInPath raises an error if the file is not found in
# the list of directories contained in path.

proc tclFileInPath {file path slave} {
    set realcheckpath [tclSafeCheckAutoPath $path $slave]
    set pwd [pwd]
    if {[file isdirectory $file]} {
	error "$file: not found"
    }
    set parent [file dirname $file]
    if {[catch {cd $parent} msg]} {
	error "$file: not found"
    }
    set realfilepath [file split [pwd]]
    foreach dir $realcheckpath {
	set match 1
	foreach a [file split $dir] b $realfilepath {
	    if {[string length $a] == 0} {
		break
	    } elseif {[string compare $a $b] != 0} {
		set match 0
		break
	    }
	}
	if $match {
	    cd $pwd
	    return 1
	}
    }
    cd $pwd
    error "$file: not found"
}

# This procedure computes our expanded copy of the path, as needed.
# It returns the path after expanding out all aliases.

proc tclSafeCheckAutoPath {path slave} {
    global auto_path
    upvar #0 tclSafe$slave state

    if {![info exists state(expanded_auto_path)]} {
	# Compute for the first time:
	set state(cached_auto_path) $path
    } elseif {"$state(cached_auto_path)" != "$path"} {
	# The value of our path changed, so recompute:
	set state(cached_auto_path) $path
    } else {
	# No change: no need to recompute.
	return $state(expanded_auto_path)
    }

    set pwd [pwd]
    set state(expanded_auto_path) ""
    foreach dir $state(cached_auto_path) {
	if {![catch {cd $dir}]} {
	    lappend state(expanded_auto_path) [pwd]
	}
    }
    cd $pwd
    return $state(expanded_auto_path)
}

proc tclSafeAliasPkgUnknown {slave package version {exact {}}} {
    tclSafeLoadPkg $slave $package $version $exact 0
}

proc tclSafeLoadPkg {slave package version exact round} {
    global auto_path tcl_policyPath tclOldPolicyPath
    upvar #0 tclSafe$slave state

    if {[string length $version] == 0} {
	set version 1.0
    }
    # Call the package init proc if it is already defined as a security policy.

    if {[tclSafeLoadPolicy $slave $package $version]} {
	return
    }

    # See if the package is a security policy.

    if {![info exists tclOldPolicyPath] ||
	    $tclOldPolicyPath != $tcl_policyPath} {

	# Search the policy path for the requested policy:

	tclSafeResearchPolicyPath

	# Try again

	if [tclSafeLoadPolicy $slave $package $version] {
	    return
	}
    }

    # The package is not a security policy, so do the regular setup.

    # Here we run tclPkgUnknown in the master, but we hijack
    # the source command so the setup ends up happening in the slave.

    set path $auto_path
    set oldpath $auto_path
    catch {set path $state(auto_path)}
    set auto_path $path

    catch {unset dir}
    rename source source.orig
    proc source {args} "upvar dir dir
	     interp eval [list $slave] tclPkgSource \[list \$dir\] \$args"
    if [catch {tclPkgUnknown $package $version $exact} err] {
	global errorInfo

	set auto_path $oldpath
	rename source {}
	rename source.orig source

	error "$err\n$errorInfo"
    }
    set auto_path $oldpath
    rename source {}
    rename source.orig source

    # If we are in the first round, check if the package
    # is now known in the slave:

    if {$round == 0} {
        set ifneeded \
		[interp eval $slave [list package ifneeded $package $version]]

	if {"$ifneeded" == ""} {

	    # Cause the policy path to be re-searched in the second round.
	    # Perhaps someone installed a policy while we were running, and
	    # we would like to notice it now.

	    unset tclOldPolicyPath
	    return [tclSafeLoadPkg $slave $package $version $exact 1]
	}
    }
}

proc tclSafeResearchPolicyPath {} {
    global tclOldPolicyPath tcl_policyPath auto_index

    set tclOldPolicyPath $tcl_policyPath

    # Loop through the path from back to front so early directories
    # end up overriding later directories.  This code is like auto_load,
    # but only new-style tclIndex files (version 2) are supported.

    for {set i [expr [llength $tcl_policyPath] - 1]} \
	    {$i >= 0} \
	    {incr i -1} {
	set dir [lindex $tcl_policyPath $i]
        set file [file join $dir tclIndex]
	if {[file exists $file]} {
	    catch {source $file}	;# Should log this error somewhere
	}
	foreach file [lsort [glob -nocomplain [file join $dir *]]] {
	    if {[file isdir $file]} {
		set dir $file
		set file [file join $file tclIndex]
		if {[file exists $file]} {
		    catch {source $file}	;# Should log this error
		}
	    }
	}
    }
}

proc tclSafeLoadPolicy {slave package version} {
    upvar #0 tclSafe$slave state
    global auto_index

    set proc ${package}_PolicyInit

    if {[info command $proc] == "$proc" ||
	    [info exists auto_index($proc)]} {
	if [info exists state(policyLoaded)] {
	    error "security policy \"$state(policyLoaded)\" already loaded"
	}
	$proc $slave $version
	interp eval $slave [list package provide $package $version]
	set state(policyLoaded) $package
	return 1
    } else {
	return 0
    }
}
# This procedure enables access from a safe interpreter to only a subset of
# the subcommands of a command:

proc tclSafeSubset {command okpat args} {
    set subcommand [lindex $args 0]
    if {[regexp $okpat $subcommand]} {
	return [eval {$command $subcommand} [lrange $args 1 end]]
    }
    error "not allowed to invoke subcommand $subcommand of $command"
}

# This procedure installs an alias in a slave that invokes "safesubset"
# in the master to execute allowed subcommands. It precomputes the pattern
# of allowed subcommands; you can use wildcards in the pattern if you wish
# to allow subcommand abbreviation.
#
# Syntax is: tclAliasSubset slave alias target subcommand1 subcommand2...

proc tclAliasSubset {slave alias target args} {
    set pat ^(; set sep ""
    foreach sub $args {
	append pat $sep$sub
	set sep |
    }
    append pat )\$
    interp alias $slave $alias {} tclSafeSubset $target $pat
}
